@@ -22,11 +22,6 @@
 #ifndef BASICIO_H
 #define BASICIO_H
 
-class SaveVecEntry;
-typedef SaveVecEntry *Handle;
-class TaskData;
-
-extern Handle IO_dispatch_c(TaskData *mdTaskData, Handle args, Handle strm, Handle code);
-extern Handle change_dirc(TaskData *mdTaskData, Handle name);
+extern struct _entrypts basicIOEPT[];
 
 #endif /* BASICIO_H */
