@@ -4,7 +4,7 @@
 
     Copyright (c) 2000-7
         Cambridge University Technical Services Limited
-    Further development Copyright David C.J. Matthews 2015-18, 2020-21.
+    Further development Copyright David C.J. Matthews 2015-18, 2020-22.
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -1435,6 +1435,51 @@ enum ByteCodeInterpreter::_returnValue ByteCodeInterpreter::RunInterpreter(TaskD
                 break;
             }
 
+            case EXTINSTR_loadPolyWord:
+            {
+                POLYUNSIGNED index = UNTAGGED(*sp++);
+                PolyObject* p = (PolyObject*)((*sp).w().AsCodePtr());
+                POLYUNSIGNED r = ((POLYUNSIGNED*)p)[index];
+                PolyObject* t = this->allocateMemory(taskData, LGWORDSIZE, pc, sp);
+                if (t == 0) goto RAISE_EXCEPTION;
+                t->SetLengthWord(LGWORDSIZE, F_BYTE_OBJ);
+                *(POLYUNSIGNED*)t = r;
+                *sp = (PolyWord)t;
+                break;
+            }
+
+            case EXTINSTR_loadNativeWord:
+            {
+                POLYUNSIGNED index = UNTAGGED(*sp++);
+                PolyObject* p = (PolyObject*)((*sp).w().AsCodePtr());
+                uintptr_t r = ((uintptr_t*)p)[index];
+                PolyObject* t = this->allocateMemory(taskData, LGWORDSIZE, pc, sp);
+                if (t == 0) goto RAISE_EXCEPTION;
+                t->SetLengthWord(LGWORDSIZE, F_BYTE_OBJ);
+                *(uintptr_t*)t = r;
+                *sp = (PolyWord)t;
+                break;
+            }
+
+            case EXTINSTR_storePolyWord:
+            {
+                uintptr_t toStore = (*(uintptr_t*)((*sp++).w().AsObjPtr()));
+                POLYUNSIGNED index = UNTAGGED(*sp++);
+                PolyObject* p = (PolyObject*)((*sp).w().AsCodePtr());
+                ((POLYUNSIGNED*)p)[index] = toStore;
+                *sp = Zero;
+                break;
+            }
+
+            case EXTINSTR_storeNativeWord:
+            {
+                uintptr_t toStore = *(uintptr_t*)((*sp++).w().AsObjPtr());
+                POLYSIGNED index = UNTAGGED(*sp++);
+                PolyObject* p = (PolyObject*)((*sp).w().AsCodePtr());
+                ((uintptr_t*)p)[index] = toStore;
+                break;
+            }
+
             case EXTINSTR_atomicExchAdd:
             {
                 // Now legacy code.
@@ -1982,7 +2027,7 @@ enum ByteCodeInterpreter::_returnValue ByteCodeInterpreter::RunInterpreter(TaskD
                     if (dv > (double)(MAXTAGGED + MAXTAGGED / 2) ||
                         dv < -(double)(MAXTAGGED + MAXTAGGED / 2))
                     {
-                        *(--sp) = (PolyWord)overflowPacket;
+                        taskData->SetException((poly_exn*)overflowPacket);
                         goto RAISE_EXCEPTION;
                     }
                     POLYSIGNED p;
@@ -2006,7 +2051,7 @@ enum ByteCodeInterpreter::_returnValue ByteCodeInterpreter::RunInterpreter(TaskD
                     // Check that the value can be tagged.
                     if (p > MAXTAGGED || p < -MAXTAGGED - 1)
                     {
-                        *(--sp) = (PolyWord)overflowPacket;
+                        taskData->SetException((poly_exn*)overflowPacket);
                         goto RAISE_EXCEPTION;
                     }
                     *sp = TAGGED(p);
